package com.fulan.application.controller;


import static org.quartz.CronScheduleBuilder.cronSchedule;
import static org.quartz.JobBuilder.newJob;
import static org.quartz.TriggerBuilder.newTrigger;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.annotation.Resource;

import org.apache.commons.lang3.StringUtils;
import org.quartz.CronExpression;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.impl.matchers.StringMatcher;
import org.quartz.impl.triggers.CronTriggerImpl;
import org.quartz.impl.triggers.SimpleTriggerImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import com.alibaba.fastjson.JSON;
import com.fulan.application.constant.JobError;
import com.fulan.application.handle.quartz.QuartzConfigration;
import com.fulan.application.util.constant.ReturnCode;
import com.fulan.application.util.date.DateUtil;
import com.fulan.application.util.domain.Response;
import com.fulan.application.util.str.StringUtil;

import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import io.swagger.annotations.Extension;
import io.swagger.annotations.ExtensionProperty;

@RestController
@RequestMapping("/job")
public class QuartzController {

    private static Logger logger = LoggerFactory.getLogger(ScheduleController.class);
    @Resource
    private Scheduler scheduler;
    @Autowired
    private QuartzConfigration quartzConfigration;

    /**
     * 定时任务查询
     *
     * @param queryJobName 任务名
     * @return
     */
    @ApiOperation(value = "定时任务查询", notes = "定时任务列表查询", response = Response.class)
    @GetMapping(value = "/list")
    public Response<Map<String, Object>> list(
            @ApiParam(name = "queryJobName", value = "根据任务名称模糊查询", required = false) @RequestParam(value = "queryJobName", required = false) String queryJobName) {
        try {
            if (StringUtils.isNotBlank(queryJobName)) {
                queryJobName = queryJobName.trim();
            }
            Integer count = 0;
            List<String> groups = scheduler.getJobGroupNames();

            // 获取正在执行的job
            Integer executingJobCount = scheduler.getCurrentlyExecutingJobs().size();
            List<HashMap<String, Object>> jobList = new ArrayList<HashMap<String, Object>>();
            Map<String, String> executingJobsMap = new HashMap<String, String>();
            for (JobExecutionContext context : scheduler.getCurrentlyExecutingJobs()) {
                executingJobsMap.put(context.getJobDetail().getKey().getGroup() + context.getJobDetail().getKey().getName(), "1");
            }

            for (String group : groups) {
                @SuppressWarnings("serial")
                Set<JobKey> jobKeys = scheduler.getJobKeys(new GroupMatcher<JobKey>(group, StringMatcher.StringOperatorName.EQUALS) {
                });

                for (JobKey jobKey : jobKeys) {

                    if (StringUtils.isNotBlank(queryJobName) && !jobKey.toString().contains(queryJobName)) {
                        continue;
                    }
                    JobDetail jobDetail = scheduler.getJobDetail(jobKey);
                    HashMap<String, Object> jobInfoMap = new HashMap<String, Object>();

                    // 触发器
                    List<? extends Trigger> triggers = scheduler.getTriggersOfJob(jobKey);
                    Map<String, Object> triggerStatusMap = new HashMap<String, Object>();
                    for (Trigger trigger : triggers) {
                        Trigger.TriggerState state = scheduler.getTriggerState(trigger.getKey());
                        triggerStatusMap.put(trigger.getKey().toString().replace(".", ""), state.name() + ":" + state.ordinal());
                        if (trigger instanceof CronTriggerImpl) {

                            jobInfoMap.put("crondesc", "cron表达式：" + ((CronTriggerImpl) trigger).getCronExpression());
                            jobInfoMap.put("preFire", DateUtil.format(trigger.getPreviousFireTime(), "yyyy-MM-dd HH:mm:ss"));
                            jobInfoMap.put("nextFire", DateUtil.format(trigger.getNextFireTime(), "yyyy-MM-dd HH:mm:ss"));

                        } else {

                            jobInfoMap.put("crondesc", "执行频率：" + ((SimpleTriggerImpl) trigger).getRepeatInterval() / 1000 + "秒");
                            jobInfoMap.put("preFire", DateUtil.format(trigger.getPreviousFireTime(), "yyyy-MM-dd HH:mm:ss"));
                            jobInfoMap.put("nextFire", DateUtil.format(trigger.getNextFireTime(), "yyyy-MM-dd HH:mm:ss"));
                        }
                    }
                    jobInfoMap.put("triggers", triggers);
                    jobInfoMap.put("jobDetail", jobDetail);
                    if (executingJobsMap.containsKey(jobDetail.getKey().getGroup() + jobDetail.getKey().getName())) {
                        jobInfoMap.put("isRunning", "正在运行");
                    } else {
                        jobInfoMap.put("isRunning", "停止");
                    }
                    jobInfoMap.put("triggerStatusMap", triggerStatusMap);
                    jobList.add(jobInfoMap);
                    count++;
                }
            }
            Map<String, Object> jobListMap = new LinkedHashMap<>();
            jobListMap.put("executingJobCount", executingJobCount);
            jobListMap.put("count", count);
            jobListMap.put("jobList", JSON.toJSONString(jobList));
            jobListMap.put("scheduler", scheduler);
            jobListMap.put("groups", groups);
            jobListMap.put("queryJobName", queryJobName);
            jobListMap.put("isValidScheduleServer", 1);
            return Response.ok(jobListMap);

        } catch (Exception e) {

            logger.error("", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);

        }

    }


    /**
     * 添加任务的处理
     */
    @ApiOperation(value = "定时任务新增", notes = "新增定时任务",
            extensions = {
                    @Extension(name = ReturnCode.API_SUB_FLAG, properties = {
                            @ExtensionProperty(name = JobError.TASK_EXIST, value = JobError.TASK_EXIST_MSG),
                            @ExtensionProperty(name = JobError.CRON_ERROR, value = JobError.CRON_ERROR_MSG),
                            @ExtensionProperty(name = JobError.JOB_CLASS_NOT_FOUND, value = JobError.JOB_CLASS_NOT_FOUND_MSG),
                    })
            }
    )
    @PostMapping(value = "/insert")
    public Response<String> insert(
            @ApiParam(name = "jobGroup", value = "任务分组", required = true) @RequestParam(value = "jobGroup") String group,
            @ApiParam(name = "jobName", value = "任务名称", required = true) @RequestParam(value = "jobName") String jobName,
            @ApiParam(name = "jobClassName", value = "任务job的class名", required = true) @RequestParam(value = "jobClassName") String jobClassName,
//            @ApiParam(name = "triggerType", value = "任务触发类型",required = true) @RequestParam(value = "triggerType") String triggerType,
//            @ApiParam(name = "rate", value = "频率") @RequestParam(value = "rate",required = false) String rate,
//            @ApiParam(name = "times", value = "次数") @RequestParam(value = "times",required = false) String times,
            @ApiParam(name = "secondField", value = "秒") @RequestParam(value = "secondField", required = false, defaultValue = "0") String second,
            @ApiParam(name = "minutesField", value = "分") @RequestParam(value = "minutesField", required = false, defaultValue = "*") String minute,
            @ApiParam(name = "hourField", value = "时") @RequestParam(value = "hourField", required = false, defaultValue = "*") String hour,
            @ApiParam(name = "dayField", value = "日") @RequestParam(value = "dayField", required = false, defaultValue = "*") String day,
            @ApiParam(name = "monthField", value = "月") @RequestParam(value = "monthField", required = false, defaultValue = "*") String month,
            @ApiParam(name = "weekField", value = "周") @RequestParam(value = "weekField", required = false, defaultValue = "?") String week,
            @ApiParam(name = "yearField", value = "年") @RequestParam(value = "yearField", required = false, defaultValue = "*") String year
    ) throws SchedulerException, IOException {

        try {
            // 验证任务是否存在
            boolean flag = scheduler.checkExists(JobKey.jobKey(jobName, group));
            if (flag) {
                return Response.fail(JobError.TASK_EXIST, JobError.TASK_EXIST_MSG);
            }

            Trigger trigger = null;
//            if ("1".equals(triggerType)) {
//
////                Integer rateInt = new Integer(rate);
////                Integer timesInt = new Integer(times);
////                trigger = newTrigger().withIdentity(jobName, group).withSchedule(simpleSchedule().withIntervalInMinutes(rateInt).withRepeatCount(timesInt).withMisfireHandlingInstructionFireNow()).build();
//            } else if ("2".equals(triggerType)) {
            String cronExpression = String.format("%s %s %s %s %s %s %s", second, minute, hour, day, month, week, year);
            boolean isValid = CronExpression.isValidExpression(cronExpression);
            if (!isValid) {
                return Response.fail(JobError.CRON_ERROR, JobError.CRON_ERROR_MSG);
            }
//                try {
//                    trigger = newTrigger().withIdentity(jobName, group).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
//                } catch (Exception e) {
//                    logger.error("addJob:", e);
//                    return new Response<>(Response.ERROR, e.getLocalizedMessage());
//                }
//            } else {
//                String cronExpression = String.format("%s %s %s %s %s %s %s", "0", "*", "*", "*", "*", "?", "2099");
//                boolean isValid = CronExpression.isValidExpression(cronExpression);
//                if (!isValid) {
//                    resp.setCode(Response.ERROR);
//                    resp.setMsg("cron表达式填写错误,您的表达式是：" + cronExpression);
//                    return resp;
//                }
            try {
                trigger = newTrigger().withIdentity(jobName, group).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
            } catch (Exception e) {
                logger.error("addJob:", e);
                return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
            }
//            }

            @SuppressWarnings("rawtypes")
            Class jobClass = null;
            try {
                jobClass = Class.forName(jobClassName);
            } catch (ClassNotFoundException e1) {
                logger.error("addJob:", e1);
            }
            if (jobClass == null) {
                return Response.fail(JobError.JOB_CLASS_NOT_FOUND, JobError.JOB_CLASS_NOT_FOUND_MSG);

            }
            JobDetail job = newJob(jobClass).withIdentity(jobName, group).build();
            scheduler.scheduleJob(job, trigger);

            // 启动
            if (!scheduler.isShutdown()) {
                scheduler.start();
            }
        } catch (Exception e) {
            logger.error("addJob:", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }

        return Response.ok();
    }


    /**
     * 中断TASK执行
     */
    @ApiOperation(value = "定时任务中断", notes = "定时任务中断", response = Response.class)
    @PostMapping(value = "/interrupt")
    public Response<String> interrupt(
            @ApiParam(name = "jobKey", value = "任务key,即任务组group+'.'+任务job名", required = true) @RequestParam("jobKey") String jobKey) throws SchedulerException {
        StringUtil.trim(jobKey);
        try {
            String group = jobKey.substring(0, jobKey.indexOf("."));
            String jobName = jobKey.substring(jobKey.indexOf(".") + 1);
            scheduler.interrupt(JobKey.jobKey(jobName, group));

            return Response.ok();
        } catch (Exception e) {
            logger.error("interruptJob:", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }
    }


    /**
     * 修改处理
     */
    @ApiOperation(value = "定时任务修改", notes = "修改定时任务",
            extensions = {
                    @Extension(name = ReturnCode.API_SUB_FLAG, properties = {
                            @ExtensionProperty(name = JobError.CRON_ERROR, value = JobError.CRON_ERROR_MSG),
                            @ExtensionProperty(name = JobError.SCHEDULE_ERROR, value = JobError.SCHEDULE_ERROR_MSG),
                    })
            }
    )
    @PostMapping(value = "/update")
    public Response<String> update(
            @ApiParam(name = "jobGroup", value = "任务分组", required = true) @RequestParam(value = "jobGroup") String group,
            @ApiParam(name = "jobName", value = "任务名称", required = true) @RequestParam(value = "jobName") String jobName,
            @ApiParam(name = "jobClassName", value = "任务job的class名", required = true) @RequestParam(value = "jobClassName") String jobClassName,
//            @ApiParam(name = "triggerType", value = "任务触发类型",required = true) @RequestParam(value = "triggerType") String triggerType,
//            @ApiParam(name = "rate", value = "频率") @RequestParam(value = "rate",required = false) String rate,
//            @ApiParam(name = "times", value = "次数") @RequestParam(value = "times",required = false) String times,
            @ApiParam(name = "secondField", value = "秒") @RequestParam(value = "secondField", required = false, defaultValue = "0") String second,
            @ApiParam(name = "minutesField", value = "分") @RequestParam(value = "minutesField", required = false, defaultValue = "*") String minute,
            @ApiParam(name = "hourField", value = "时") @RequestParam(value = "hourField", required = false, defaultValue = "*") String hour,
            @ApiParam(name = "dayField", value = "日") @RequestParam(value = "dayField", required = false, defaultValue = "*") String day,
            @ApiParam(name = "monthField", value = "月") @RequestParam(value = "monthField", required = false, defaultValue = "*") String month,
            @ApiParam(name = "weekField", value = "周") @RequestParam(value = "weekField", required = false, defaultValue = "?") String week,
            @ApiParam(name = "yearField", value = "年") @RequestParam(value = "yearField", required = false, defaultValue = "*") String year)
            throws SchedulerException {
        try {

            List<? extends Trigger> triggers = scheduler.getTriggersOfJob(JobKey.jobKey(jobName, group));
            jobName = triggers.get(0).getKey().getName();
            group = triggers.get(0).getKey().getGroup();
            Trigger trigger = null;
//            if ("1".equals(triggerType)) {
//
//                Integer rateInt = new Integer(rate);
//                Integer timesInt = new Integer(times);
//                trigger = newTrigger().withIdentity(jobName, group).withSchedule(simpleSchedule().withIntervalInMinutes(rateInt).withRepeatCount(timesInt).withMisfireHandlingInstructionFireNow()).build();
//            } else if ("2".equals(triggerType)) {
            String cronExpression = String.format("%s %s %s %s %s %s %s", second, minute, hour, day, month, week, year);
            boolean isValid = CronExpression.isValidExpression(cronExpression);
            if (!isValid) {
                return Response.fail(JobError.CRON_ERROR, JobError.CRON_ERROR_MSG);
            }
//                try {
//                    trigger = newTrigger().withIdentity(jobName, group).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
//                } catch (Exception e) {
//                    logger.error("newTrigger Exception",e);
//
//                    return new Response<>(Response.ERROR, e.getLocalizedMessage());
//
//                }
//            } else {
//                String cronExpression = String.format("%s %s %s %s %s %s %s", "0", "*", "*", "*", "*", "?", "2099");
//                boolean isValid = CronExpression.isValidExpression(cronExpression);
//                if (!isValid) {
//
//                    resp.setCode(Response.ERROR);
//                    resp.setMsg("cron表达式填写错误,您的表达式是：" + cronExpression);
//                    return resp;
//
//                }
            try {
                trigger = newTrigger().withIdentity(jobName, group).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
            } catch (Exception e) {
                logger.error("newTrigger Exception", e);
                return Response.fail(JobError.SCHEDULE_ERROR, JobError.SCHEDULE_ERROR_MSG);
            }
//            }
            scheduler.rescheduleJob(trigger.getKey(), trigger);
        } catch (Exception e) {
            logger.error("newTrigger Exception", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }

        return Response.ok();
    }

    /**
     * 删除TASK
     */
    @ApiOperation(value = "定时任务删除", notes = "删除定时任务", response = Response.class)
    @PostMapping(value = "/delete")
    public Response<String> delete(
            @ApiParam(name = "jobKey", value = "任务key,即任务组group+'.'+任务job名", required = true) @RequestParam(value = "jobKey", required = true) String jobKey) throws SchedulerException {
        try {

            jobKey = jobKey.trim();
            String group = jobKey.substring(0, jobKey.indexOf("."));
            String jobName = jobKey.substring(jobKey.indexOf(".") + 1);
            scheduler.deleteJob(JobKey.jobKey(jobName, group));

            return Response.ok();
        } catch (Exception e) {
            logger.error("deleteJob:", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }

    }

    /**
     * 执行某个TASK一次
     */
    @ApiOperation(value = "定时任务执行", notes = "将一个定时任务执行一次", response = Response.class)
    @PostMapping(value = "/executeOnce")
    public Response<String> executeOnce(@ApiParam(name = "jobKey", value = "任务key,即任务组group+'.'+任务job名", required = true) @RequestParam(value = "jobKey", required = true) String jobKey) throws SchedulerException {

        StringUtil.trim(jobKey);
        try {
            String group = jobKey.substring(0, jobKey.indexOf("."));
            String jobName = jobKey.substring(jobKey.indexOf(".") + 1);
            Trigger trigger = newTrigger().withIdentity(jobName + UUID.randomUUID().toString() + "MANUALLY", group).withPriority(100).forJob(JobKey.jobKey(jobName, group)).build();
            scheduler.scheduleJob(trigger);

            return Response.ok();
        } catch (SchedulerException e) {
            logger.error("executeOnce:", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }
    }


    /**
     * 定时任务jobClassList
     */
    @ApiOperation(value = "定时任务jobClassList", notes = "定时任务jobClassList", response = Response.class)
    @GetMapping(value = "/class/list")
    public Response<Map<String, List<String>>> classList() throws SchedulerException {

        try {
            // 存在的任务类
            Map<String, List<String>> jobClassesMap = quartzConfigration.getJobClassesMap();
            return Response.ok(jobClassesMap);
        } catch (Exception e) {
            logger.error("jobClassList:", e);
            return Response.fail(ReturnCode.SERVICE_UNAVAILABLE);
        }


    }



}
