package com.fulan.application.controller.remote;


import static org.quartz.CronScheduleBuilder.cronSchedule;
import static org.quartz.JobBuilder.newJob;
import static org.quartz.TriggerBuilder.newTrigger;

import java.io.IOException;
import java.util.List;

import javax.annotation.Resource;

import org.quartz.CronExpression;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import com.fulan.application.constant.JobError;
import com.fulan.application.exception.BizException;
import com.fulan.application.handle.quartz.QuartzConfigration;
import com.fulan.application.job.domain.VmyJob;
import com.fulan.application.util.constant.ReturnCode;
import com.fulan.application.util.domain.Response;

import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import io.swagger.annotations.Extension;
import io.swagger.annotations.ExtensionProperty;

@RestController
@RequestMapping("/remote/job")
public class RemoteQuartzController {

    @Resource
    private Scheduler scheduler;
    @Autowired
    private QuartzConfigration quartzConfigration;


    private static Logger logger = LoggerFactory.getLogger(RemoteQuartzController.class);

    /**
     * 添加任务的处理
     */
    @ApiOperation(value = "定时任务新增", notes = "新增定时任务",
    		extensions = {
				       @Extension( name = ReturnCode.API_SUB_FLAG, properties = {
				               @ExtensionProperty(name = JobError.TASK_EXIST, value = JobError.TASK_EXIST_MSG),
				               @ExtensionProperty(name = JobError.CRON_ERROR, value = JobError.CRON_ERROR_MSG),
				               @ExtensionProperty(name = JobError.JOB_CLASS_NOT_FOUND, value = JobError.JOB_CLASS_NOT_FOUND_MSG),
				           })
                }    
    		)
    @PostMapping(value = "/insert")
    public void insert(@RequestBody VmyJob job) throws SchedulerException, IOException {
    	// 验证任务是否存在
        boolean flag = scheduler.checkExists(JobKey.jobKey(job.getJobName(), job.getGroup()));
        if (flag) {
        	throw new BizException(JobError.TASK_EXIST,JobError.TASK_EXIST_MSG);
        }

        Trigger trigger = null;
        String cronExpression = String.format("%s %s %s %s %s %s %s", 
        		job.getSecond(), job.getMinute(), job.getHour(), job.getDay(), job.getMonth(), job.getWeek(), job.getYear());
        boolean isValid = CronExpression.isValidExpression(cronExpression);
        if (!isValid) {
        	throw new BizException(JobError.CRON_ERROR,JobError.CRON_ERROR_MSG);
        }
        try {
            trigger = newTrigger().withIdentity(job.getJobName(), job.getGroup()).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
        } catch (Exception e) {
            logger.error("addJob:", e);
            throw new BizException(ReturnCode.SERVICE_UNAVAILABLE.getCode(),ReturnCode.SERVICE_UNAVAILABLE.getMsg());
        }

        @SuppressWarnings("rawtypes")
        Class jobClass = null;
        try {
            jobClass = Class.forName(job.getJobClassName());
        } catch (ClassNotFoundException e1) {
            logger.error("addJob:", e1);
        }
        if (jobClass == null) {
        	throw new BizException(JobError.JOB_CLASS_NOT_FOUND,JobError.JOB_CLASS_NOT_FOUND_MSG);

        }
        JobDetail jobDetail = newJob(jobClass).withIdentity(job.getJobName(), job.getGroup()).build();
        scheduler.scheduleJob(jobDetail, trigger);

        // 启动
        if (!scheduler.isShutdown()) {
            scheduler.start();
        }
    }

    /**
     * 修改处理
     */
    @ApiOperation(value = "定时任务修改", notes = "修改定时任务", 
    		extensions = {
				       @Extension( name = ReturnCode.API_SUB_FLAG, properties = {
				               @ExtensionProperty(name = JobError.CRON_ERROR, value = JobError.CRON_ERROR_MSG),
				               @ExtensionProperty(name = JobError.SCHEDULE_ERROR, value = JobError.SCHEDULE_ERROR_MSG),
				           })
             }  
    		)
    @PostMapping(value = "/update")
    public void update(@RequestBody VmyJob job)throws SchedulerException {
    	String jobName=job.getJobName();
    	String group=job.getGroup();
        List<? extends Trigger> triggers = scheduler.getTriggersOfJob(JobKey.jobKey(jobName, group));
        jobName = triggers.get(0).getKey().getName();
        group = triggers.get(0).getKey().getGroup();
        Trigger trigger = null;
        String cronExpression = String.format("%s %s %s %s %s %s",
        		job.getSecond(), job.getMinute(), job.getHour(), job.getDay(), job.getMonth(), job.getWeek(), job.getYear());
        boolean isValid = CronExpression.isValidExpression(cronExpression);
        if (!isValid) {
        	new BizException(JobError.CRON_ERROR,JobError.CRON_ERROR_MSG);
        }
        try {
            trigger = newTrigger().withIdentity(jobName, group).withSchedule(cronSchedule(cronExpression).withMisfireHandlingInstructionFireAndProceed()).build();
        } catch (Exception e) {
            logger.error("newTrigger Exception",e);
            new BizException(JobError.SCHEDULE_ERROR,JobError.SCHEDULE_ERROR_MSG);
        }
        scheduler.rescheduleJob(trigger.getKey(), trigger);
    }

    /**
     * 删除TASK
     */
    @ApiOperation(value = "定时任务删除", notes = "删除定时任务", response = Response.class)
    @PostMapping(value = "/delete")
    public void delete(
            @ApiParam(name = "jobKey", value = "任务key,即任务组group+'.'+任务job名",required = true) @RequestParam(value = "jobKey",required = true) String jobKey) throws SchedulerException {
    	jobKey = jobKey.trim();
        String group = jobKey.substring(0, jobKey.indexOf("."));
        String jobName = jobKey.substring(jobKey.indexOf(".")+1);
        scheduler.deleteJob(JobKey.jobKey(jobName, group));
    }
}
